/* Output from p2c, the Pascal-to-C translator */
/* From input file "netsim.pas" */


#include "decl.h"
#include <stdlib.h>
#include <stdio.h>

#define NETSIM_G
#include "netsim.h"


typedef struct enode {
  float idx;
  struct enode *skip, *next, *prev;
  erec *e;
} enode;


static float current_time;
static enode *eventlist;
static boolean save_events;
static enode dummy_node;
static erec dummy_event;


/* AHA function to return time under win32 */

void gettime(time_t *pointer)
{
	time_t long_time;
	time( &long_time );                /* Get time as long integer. */
	*pointer = long_time;
}



static void einsert(erec *x, enode *current)
{
  enode *temp;

  temp = (enode *)malloc(sizeof(enode));
  /* link in new node to previous and next nodes */
  temp->prev = current->prev;
  temp->next = current;

  /* set indexing fields */
  temp->skip = current;
  temp->idx = current->e->time;

  /* set event pointer */
  temp->e = x;

  /* fix previous node to point to new node,
     unless current is present head of list */
  if (current->prev != NULL)
    current->prev->next = temp;

  /* fix next node to point to new node */
  current->prev = temp;

  /* inserted at head of list, replace eventlist */
  if (current == eventlist)
    eventlist = temp;

  /* inserted at tail of list, add shortcut to prevent
     O( N^2 ) performance for (nearly) ordered input by
     always having first node's index point to last node */
  if (current == &dummy_node && eventlist != temp) {
    /*NOTE: above also prevents loop when only one node on list */
    eventlist->skip = temp;
    eventlist->idx = temp->e->time;
  }
}


void queue_event(erec *e)
{
  enode *current;

	/* AHA debug 
  if(e->time > 2.6)
	  printf("Event with e->time=%f\n", e->time);*/


  if (e->time < current_time) 
  {
    printf(
      "%%NETSIM - Event with illegal time stamp [% .5E] received by Queue_Event.\n",
      e->time);
    return;
  }

  current = eventlist;
  while (current->next != NULL) 
  {   /*not at end of list... */
    if (e->time >= current->idx)   /* use skip pointer to bypass N nodes */
		current = current->skip;
    else 
		{  /* check the node's event time value */
		if (e->time < current->e->time) 
			{   /* not found yet... */
			einsert(e, current);
			return;
			}
		current = current->next;
		}
  }
  einsert(e, current);

  /* proper place to insert new node */
}


erec *next_event()
{
  erec *Result;

  do 
  {
    current_time = eventlist->e->time;
    Result = eventlist->e;
    eventlist = eventlist->next;
    if (!save_events) 
	{
		if (eventlist != NULL) 
		{
			Free(eventlist->prev);
  			eventlist->prev = NULL;
		}
    }
  } while (current_time < 0.0);
  /*i.e., non-cancelled event,
                             dummy event always triggers this condition*/

  if (eventlist == NULL) 
  {
    /* was at end of list, just read dummy event,
                              return NIL event pointer */
    Result = NULL;
    eventlist = &dummy_node;   /*reset eventlist to point to 'empty' list*/
  }
  return Result;
}


void cancel_events(long src)
{
  enode *current;

  current = eventlist;
  while (current != &dummy_node) {   /*at end of list... */
    if (current->e->src == src)   /* cancel this event by negating time */
      current->e->time = -current->e->time;
    current = current->next;
  }
}


float sim_time()
{
  return current_time;
}


void set_save_mode()
{
  save_events = true;
}


void set_event_tracking(symbol_node *options)
{
}


/* p2c: Note: Remember to call _netsim_init() in main program */

void _netsim_init()
{
  static int _was_initialized = 0;
  if (_was_initialized++)
    return;
  dummy_event.time = (float)1.0e30;
  dummy_event.next = -1;
  dummy_event.src = -1;
  dummy_event.dst = -1;
  dummy_event.msg = NULL;

  dummy_node.idx = (float)1.0e30;
  dummy_node.next = NULL;
  dummy_node.prev = NULL;
  dummy_node.skip = NULL;
  dummy_node.e = &dummy_event;

  eventlist = &dummy_node;
  current_time = (float)0.0;
  save_events = false;
  /* AHA debug 
  printf(" _netsim_init() passed\n");*/
}


/* End. */
